/***************************************************************************
                     MPEG_audio_module.h  -  description
                             -------------------
   MPEG audio PES module functions.
                             -------------------
    begin                : Thu Apr 14 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define MPEG_AUDIO_MODULE_TRACE*/
#ifdef MPEG_AUDIO_MODULE_TRACE
# include <stdio.h>
#endif
/*#include <string.h>*/

#include "PES_PKT.h"
#include "commands.h"
#include "sound/sound.h"
#include "MPEG_audio_module.h"
#include "MPEG_audio.h"
#include "layer2_data.h"

typedef struct {
   MPEG_AUDIO_FRAME frame;
   LAYER2_DATA      layer;
   u16              sound_init, frame_bytes;
   s32              frame_interval;
   SYS_DATA        *sys_data;
} MPEG_audio_module_data;

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t MPEG_AUDIO_MODULE_mem_needed()
{
   return sizeof(MPEG_audio_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void MPEG_AUDIO_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   MPEG_audio_module_data *data = (MPEG_audio_module_data *)mem;

   data->frame_interval = 0;
   data->frame.state    = MPEG_AUDIO_next_frame_STATE;

   if (sys_data) {
/*      if (!data->sys_data) {
      printf("data               =0x%08X\n"
             "data->frame        =0x%08X\n"
             "data->layer        =0x%08X\n"
             "layer->sample_array=0x%08X 0x%08X\n",
             (u32)data,
             (u32)&(data->frame),
             (u32)&(data->layer),
             (u32)&(data->layer.sample_array0),
             (u32)&(data->layer.sample_array1));
      }*/
      data->sys_data      = sys_data;
      sys_data->a_sync    = 0;
      data->frame.a_sync  = &(sys_data->a_sync);
      data->frame.pes_pts = 0;
   }
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void MPEG_AUDIO_MODULE_done(void *mem)
{
   MPEG_audio_module_data *data = (MPEG_audio_module_data *)mem;
   
   if (data->sound_init) {
      data->sound_init = 0;
      SOUND_done();
   }

   data->frame_interval = 0;
   data->frame.state    = MPEG_AUDIO_next_frame_STATE;

   if (data->sys_data) {
      data->sys_data->a_sync = 0;
      data->frame.pes_pts    = 0;
   }
}

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void MPEG_AUDIO_MODULE_set_offset(void *mem, CLK_VAL offset)
{
   MPEG_audio_module_data *data = (MPEG_audio_module_data *)mem;
   
   data->sys_data->a_sync = 0;
   data->frame.offset     = offset;
   data->frame.pes_pts    = 0;
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void MPEG_AUDIO_MODULE_discontinuity(void *mem)
{
   MPEG_audio_module_data *data = (MPEG_audio_module_data *)mem;

   data->frame.state = MPEG_AUDIO_next_frame_STATE;
   if (data->sys_data) {
      data->sys_data->a_sync = 0;
      data->frame.pes_pts    = 0;
   }
}


/**************************************************************************
   PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the  PES_PKTs
       they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
int MPEG_AUDIO_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx)
{
   return (pmt->elem[idx]->stream_type == STREAM_TYPE_MPEG1_AUDIO);
}

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void MPEG_AUDIO_MODULE_process_pkt(void *mem, void *pes)
{
   u32                     bytes_read;
   MPEG_audio_module_data *data = (MPEG_audio_module_data *)mem;
   PES_PKT                *pkt  = (PES_PKT *)pes;

#ifndef _SPEED_TEST_
   if (PES_PTS_FLAG(pkt))
      data->frame.pes_pts = (pkt->PTS) ? pkt->PTS : 1;
#endif

#ifdef MPEG_AUDIO_MODULE_TRACE
   printf("start MPEG_AUDIO_MODULE_process_pkt()\n");
#endif
   while (pkt->payload_len16) {
      if (data->frame.state < MPEG_AUDIO_next_frame_END_STATE) {
         /* look for a new audio frame */
         bytes_read = MPEG_AUDIO_next_frame(pkt->payload,
                                            (u32)pkt->payload_len16,
                                            &(data->frame));
         if ( (data->frame.state == MPEG_AUDIO_next_frame_END_STATE)&&
              (data->frame.frame_len > 4) ) {
            data->frame_bytes = (u16)(data->frame.frame_len - 4);
            if ( (!data->frame_interval)&&(data->frame.sample_rate) ) {
               data->frame_interval = (s32)((36*32*MPEG2_SYS_CLK_FREQ)/
                                            data->frame.sample_rate);
            }
            data->sys_data->a_pos += (CLK_VAL)data->frame_interval;
            
            if ( (data->sys_data->a_skip > 0)||
                 (data->sys_data->command == CMD_SLOW_FWD) ) {
               if (data->sys_data->a_skip > 0)
                  data->sys_data->a_skip -= (CLK_VAL)data->frame_interval;
               data->sys_data->a_sync = 0;
               data->frame.state      = LAYER2_read_END_STATE; /* don't play audio */
               SOUND_stop(0);
            }
            else if (!data->sound_init) {
               if (SOUND_init((u32)data->frame.num_ch, data->frame.sample_rate)) {
                  data->layer.audio_buf = SOUND_buf(36*32);
                  if (data->layer.audio_buf)
                     data->sound_init = 1;
               }
            }
#ifdef MPEG_AUDIO_MODULE_TRACE
            printf("<START_FRAME>\n");
#endif
         }
         else
            bytes_read = (u32)pkt->payload_len16;
      }
      else {
#ifdef MPEG_AUDIO_MODULE_TRACE
         printf("  state=0x%04hX  bytes=%hu\n", 
                data->frame.state, data->frame_bytes);
#endif
         /* deal with the payload */
         if (pkt->payload_len16 >= data->frame_bytes) {
            bytes_read         = (u32)data->frame_bytes + 1;
            data->frame_bytes  = 0; /* this is the end of the frame */
         }
         else {
            bytes_read         = (u32)pkt->payload_len16;
            data->frame_bytes -= pkt->payload_len16;
         }
         
         if ( (data->sound_init)&&(data->frame.state < LAYER2_read_END_STATE) ) {
            data->frame.payload       = pkt->payload;
            data->frame.payload_len16 = (u16)bytes_read;
            LAYER2_read(&(data->frame), &(data->layer));
         }
         
         if (!data->frame_bytes) {
            data->frame.state = MPEG_AUDIO_next_frame_STATE;
            bytes_read--;
         }
#ifdef MPEG_AUDIO_MODULE_TRACE
         printf("  state=0x%04hX  bytes=%hu\n", 
                data->frame.state, data->frame_bytes);
         if (!data->frame_bytes)
            printf("<END___FRAME>\n\n");
#endif
      }

      pkt->payload_len16 -= (u16)bytes_read;
      pkt->payload       += bytes_read;
   }
#ifdef MPEG_AUDIO_MODULE_TRACE
   printf("end MPEG_AUDIO_MODULE_process_pkt()\n\n");
#endif
}


