/***************************************************************************
                        layer2_data.c  -  description
                             -------------------
   Functions for handling layer 2 data.
                             -------------------
    begin                : Thu Apr 19 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define _MP2_TRACE_*/
#ifdef _MP2_TRACE_
#include <stdio.h>
#endif
#include <string.h>

#include "sound/sound.h"
#include "layer2_data.h"
#include "synth.h"
#include "bitstream.h"

#define DATA_BYTE       layer->data
#define BIT_SHIFT_BYTE  layer->bit_shift
#define DATA_PTR        ptr
#define END_DATA_PTR    end
#define PARSER_STATE    frame->state
#define STOP_PARSING    return


#define SF1_SCALE(qb)      (((s64)1 << (qb+FP_SHIFT)) / (((s64)1 << (qb)) - (s64)1))
#define SF1_CONST(c, qb)   (s32)(((s64)c * SF1_SCALE(qb)) >> FP_SHIFT)
#define SF1_ROW(qb)        SF1_CONST(FP_CONST(2.0), qb), \
                           SF1_CONST(FP_CONST(1.587401052), qb), \
                           SF1_CONST(FP_CONST(1.25992105), qb)
#define SF2_ROW(den)       (AV1)FP_CONST(4.0/den), \
                           (AV1)FP_CONST(3.174802104/den), \
                           (AV1)FP_CONST(2.5198421/den)


#ifdef _MP2_TRACE_
static int frame_cnt = 0;
#endif
static const u8  alloc_table[32]  = {4,4,4,4,4,4,4,4,4,4,4,
                                     3,3,3,3,3,3,3,3,3,3,3,3,
                                     2,2,2,2,2,2,2, 0,0},
                                   /*    1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15*/
                 qidx_table[6*16] = { 2, 0, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,
                                     10, 0, 1, 3, 2, 4, 5, 6, 7, 8, 9,10,11,12,13,16,
                                     22, 0, 1, 3, 2, 4, 5,16,16,16,16,16,16,16,16,16,
                                     99, 0, 1,16,16,16,16,16,16,16,16,16,16,16,16,16,
                                      1, 0, 1, 2, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
                                     99, 0, 1, 2, 4, 5, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7},

                 qbits_table[4]   = {5, 7,10, 0},
                 qsteps_table[4]  = {3, 5, 9, 0};

static const s32 sf_table1[15][3] = {{SF1_ROW( 2)}, {SF1_ROW( 3)}, {SF1_ROW( 4)},
                                     {SF1_ROW( 5)}, {SF1_ROW( 6)}, {SF1_ROW( 7)},
                                     {SF1_ROW( 8)}, {SF1_ROW( 9)}, {SF1_ROW(10)},
                                     {SF1_ROW(11)}, {SF1_ROW(12)}, {SF1_ROW(13)},
                                     {SF1_ROW(14)}, {SF1_ROW(15)}, {SF1_ROW(16)}},
                 sf_table2[3][3]  = {{SF2_ROW( 3)}, {SF2_ROW( 5)}, {SF2_ROW( 9)}};

static void requant_samples(AV1 *sample, u8 qbits, u8 scalefactor);
static void requant_group(AV1 *sample, u8 qsteps, u8 scalefactor);


/* Processes the layer 2 data in the MPEG_AUDIO_FRAME.
      INPUT: frame - Points to the MPEG_AUDIO_FRAME holding the
                       layer 2 data. This frame does not have
                       to contain all the data. Repeated calls to
                       this function can be used to read the data
                       in parts.
      OUTPUT: layer - Points to the LAYER2_DATA recovered from the
                        MPEG_AUDIO_FRAME. */
void LAYER2_read(MPEG_AUDIO_FRAME *frame,
                 LAYER2_DATA      *layer)
{
   s16       audio[32],
            *audio_buf;
   const u8 *ptr = frame->payload,
            *end = ptr + frame->payload_len16;
   u8        bit_shift_inc;

#define ALLOC_1         (LAYER2_read_STATE+ 1)
#define ALLOC_2         (LAYER2_read_STATE+ 2)
#define SCFSI           (LAYER2_read_STATE+ 3)
#define SCALEFACTOR_1   (LAYER2_read_STATE+ 4)
#define SCALEFACTOR_2   (LAYER2_read_STATE+ 5)
#define SCALEFACTOR_3   (LAYER2_read_STATE+ 6)
#define SAMPLE_1        (LAYER2_read_STATE+ 7)
#define SAMPLE_2        (LAYER2_read_STATE+ 8)
#define SAMPLE_3        (LAYER2_read_STATE+ 9)
#define SAMPLE_4        (LAYER2_read_STATE+10)
#define SAMPLE_5        (LAYER2_read_STATE+11)
#define SAMPLE_6        (LAYER2_read_STATE+12)
#define SAMPLE_7        (LAYER2_read_STATE+13)
#define SAMPLE_8        (LAYER2_read_STATE+14)

   if (frame->state > LAYER2_read_STATE) {
      switch (frame->state) {
      case ALLOC_1:
         goto READ_ALLOC_1;
      case ALLOC_2:
         goto READ_ALLOC_2;
      case SCFSI:
         goto READ_SCFSI;
      case SCALEFACTOR_1:
         goto READ_SCALEFACTOR_1;
      case SCALEFACTOR_2:
         goto READ_SCALEFACTOR_2;
      case SCALEFACTOR_3:
         goto READ_SCALEFACTOR_3;
      case SAMPLE_1:
         goto READ_SAMPLE_1;
      case SAMPLE_2:
         goto READ_SAMPLE_2;
      case SAMPLE_3:
         goto READ_SAMPLE_3;
      case SAMPLE_4:
         goto READ_SAMPLE_4;
      case SAMPLE_5:
         goto READ_SAMPLE_5;
      case SAMPLE_6:
         goto READ_SAMPLE_6;
      case SAMPLE_7:
         goto READ_SAMPLE_7;
      case SAMPLE_8:
         goto READ_SAMPLE_8;
      default:
#ifdef _MP2_TRACE_
         printf("LAYER2_read_ERROR\n\n");
#endif
         goto LAYER2_read_DONE;
      }
   }
   else {
#ifdef _MP2_TRACE_
/*      if (!frame_cnt) {
         int i;
         printf("scale_factor1:\n");
         for (i = 0; i < 15; i++) {
            printf("  %2d> 0x%08X 0x%08X 0x%08X\n",
                  i,
                  sf_table1[i][0],
                  sf_table1[i][1],
                  sf_table1[i][2]);
         }
         printf("scale_factor2:\n");
         for (i = 0; i < 3; i++) {
            printf("  %2d> 0x%08X 0x%08X 0x%08X\n",
                  i,
                  sf_table2[i][0],
                  sf_table2[i][1],
                  sf_table2[i][2]);
         }
         printf("\n");
      }*/
      printf("START DECODE LAYER 2 (FRAME %d)\n", frame_cnt);
#endif
      layer->bit_shift    = 8;
      layer->alloc_bits   = alloc_table;
      layer->qindex_table = qidx_table;
      if (frame->sblimit <= 12) {
         layer->alloc_bits   += 9;
         layer->qindex_table += 4*16;
      }
   }

   /***** READ ALLOC *****/
#ifdef _MP2_TRACE_
   printf("\n  alloc[]=\n");
#endif
   bit_shift_inc = 0;
   layer->alloc  = layer->alloc_array;
   for (layer->sb = 0; layer->sb < frame->bound; layer->sb++) {
#ifdef _MP2_TRACE_
      printf("    %2hu>", (u16)layer->sb);
#endif
      for (layer->ch = 0; layer->ch < frame->num_ch; layer->ch++) {
         ALIGN_DATA_BYTE(ALLOC_1, bit_shift_inc);
READ_ALLOC_1:
         GET_LSB;
         bit_shift_inc   = *(layer->alloc_bits);
         *(layer->alloc) = DATA_BYTE >> (8 - bit_shift_inc);
#ifdef _MP2_TRACE_
         printf(" %2hu", (u16)*(layer->alloc));
#endif
         layer->alloc++;
      }
      layer->alloc_bits++;
#ifdef _MP2_TRACE_
      printf("\n");
#endif
   }
   for (; layer->sb < frame->sblimit; layer->sb++) {
      ALIGN_DATA_BYTE(ALLOC_2, bit_shift_inc);
READ_ALLOC_2:
      GET_LSB;
      bit_shift_inc     = *(layer->alloc_bits++);
      *(layer->alloc)   = 
      *(layer->alloc+1) = DATA_BYTE >> (8 - bit_shift_inc);
#ifdef _MP2_TRACE_
      printf("    %2hu> %2hu\n", (u16)layer->sb, (u16)*(layer->alloc));
#endif
      layer->alloc += 2;
   }

   /***** READ SCFSI *****/
#ifdef _MP2_TRACE_
   printf("\n  scfsi[]=\n");
#endif
   layer->alloc = layer->alloc_array;
   layer->scfsi = layer->scfsi_array;
   for (layer->sb = 0; layer->sb < frame->sblimit; layer->sb++) {
#ifdef _MP2_TRACE_
      printf("    %2hu>", (u16)layer->sb);
#endif
      for (layer->ch = 0; layer->ch < frame->num_ch; layer->ch++) {
         if (*(layer->alloc++)) {
            ALIGN_DATA_BYTE(SCFSI, bit_shift_inc);
READ_SCFSI:
            bit_shift_inc   = 2;
            GET_LSB;
            *(layer->scfsi) = DATA_BYTE >> (8 - 2);
#ifdef _MP2_TRACE_
            printf(" %hu", (u16)*(layer->scfsi));
#endif
         }
         layer->scfsi++;      
      }
#ifdef _MP2_TRACE_
      printf("\n");
#endif
   }

   /***** READ SCALEFACTOR *****/
#ifdef _MP2_TRACE_
   printf("\n  scalefactor[]=\n");
#endif
   layer->alloc       = layer->alloc_array;
   layer->scfsi       = layer->scfsi_array;
   layer->scalefactor = layer->scalefactor_array;
   for (layer->sb = 0; layer->sb < frame->sblimit; layer->sb++) {
#ifdef _MP2_TRACE_
      printf("    %2hu>", (u16)layer->sb);
#endif
      for (layer->ch = 0; layer->ch < frame->num_ch; layer->ch++) {
         if (*(layer->alloc++)) {
            ALIGN_DATA_BYTE(SCALEFACTOR_1, bit_shift_inc);
READ_SCALEFACTOR_1:
            GET_LSB;
            *(layer->scalefactor) = DATA_BYTE >> (8 - 6);

            if ( (*(layer->scfsi) == 0)||
                 (*(layer->scfsi) == 3) ) {
               ALIGN_DATA_BYTE(SCALEFACTOR_2, 6);
READ_SCALEFACTOR_2:
               GET_LSB;
               *(layer->scalefactor+1) = DATA_BYTE >> (8 - 6);
            }
            else
               *(layer->scalefactor+1) = *(layer->scalefactor);

            if (*(layer->scfsi) < 2) {
               ALIGN_DATA_BYTE(SCALEFACTOR_3, 6);
READ_SCALEFACTOR_3:
               GET_LSB;
               *(layer->scalefactor+2) = DATA_BYTE >> (8 - 6);
            }
            else
               *(layer->scalefactor+2) = *(layer->scalefactor+1);

#ifdef _MP2_TRACE_
            printf(" %2hu %2hu %2hu  ", (u16)*(layer->scalefactor),
                                        (u16)*(layer->scalefactor+1),
                                        (u16)*(layer->scalefactor+2));
#endif
            
            bit_shift_inc = 6;
         }
         layer->scfsi++;
         layer->scalefactor += 3;
      }
#ifdef _MP2_TRACE_
      printf("\n");
#endif
   }

   /***** READ SAMPLES *****/
#ifdef _MP2_TRACE_
   printf("\n");
#endif
   layer->alloc_bits = layer->qindex_table;
   memset(layer->sample_array0, 0, (sizeof(AV1)*36*32));
   memset(layer->sample_array1, 0, (sizeof(AV1)*36*32));
   
   for (layer->gr = 0; layer->gr < 12; layer->gr++) {
      layer->qindex_table = layer->alloc_bits;
      layer->alloc        = layer->alloc_array;
      layer->scalefactor  = layer->scalefactor_array + ((u32)layer->gr >> 2);
      layer->sample_gr    = layer->sample_array0 + ((u32)layer->gr * 32*3);
      
      for (layer->sb = 0; layer->sb < frame->bound; layer->sb++) {
         layer->sample = layer->sample_gr;
         
         for (layer->ch = 0; layer->ch < frame->num_ch; layer->ch++) {
            if (*(layer->alloc)) {
               /* get sample values sample[0], sample[32], sample[64] */
               layer->qindex = layer->qindex_table[*(layer->alloc)];
               if (layer->qindex < 3) {
                  /* read the grouped sample code */
                  layer->qbits = qbits_table[layer->qindex];
                  if (layer->qbits > 8) {
                     layer->qbits -= 8;
                     ALIGN_DATA_BYTE(SAMPLE_1, bit_shift_inc);
READ_SAMPLE_1:
                     GET_LSB;
                     *(layer->sample) = (AV1)DATA_BYTE << layer->qbits;
                     bit_shift_inc    = 8;
                  }
                  ALIGN_DATA_BYTE(SAMPLE_2, bit_shift_inc);
READ_SAMPLE_2:
                  GET_LSB;
                  *(layer->sample) |= (AV1)DATA_BYTE >> (8 - layer->qbits);
                  bit_shift_inc     = layer->qbits;
#ifdef _MP2_TRACE_
                  printf("     value: %d", *(layer->sample));
#endif
                  requant_group(layer->sample, 
                                qsteps_table[layer->qindex],
                                *(layer->scalefactor));
               }
               else {
                  /* read the 3 separate samples */
#ifdef _MP2_TRACE_
                  printf("    values:");
#endif
                  for (layer->s = 0; layer->s < 3; layer->s++) {
                     layer->qbits = layer->qindex;
                     if (layer->qbits > 8) {
                        layer->qbits -= 8;
                        ALIGN_DATA_BYTE(SAMPLE_3, bit_shift_inc);
READ_SAMPLE_3:
                        GET_LSB;
                        *(layer->sample) = (AV1)DATA_BYTE << layer->qbits;
                        bit_shift_inc    = 8;
                     }
                     ALIGN_DATA_BYTE(SAMPLE_4, bit_shift_inc);
READ_SAMPLE_4:
                     GET_LSB;
                     *(layer->sample) |= (AV1)DATA_BYTE >> (8 - layer->qbits);
#ifdef _MP2_TRACE_
                     printf(" %d", *(layer->sample));
#endif
                     layer->sample    += 32;    /* move to next sample */
                     bit_shift_inc     = layer->qbits;
                  }
                  layer->sample -= 3*32;        /* move back to first sample */
                  requant_samples(layer->sample, 
                                  layer->qindex, 
                                  *(layer->scalefactor));
               }
#ifdef _MP2_TRACE_
               printf("\n  sample[%hu][%2hu][%2hu]=%d, %d, %d\n",
                      (u16)layer->ch, (u16)(layer->gr*3), (u16)layer->sb,
                      *(layer->sample), *(layer->sample+32), *(layer->sample+64));
#endif
            }
            layer->alloc++;
            layer->scalefactor += 3;            /* move to next channel */
            layer->sample      += (15+36)*32;   /* move to next channel */
         }
         layer->sample_gr++;                    /* move to next sub-band */
         if (layer->sb == *(layer->qindex_table))
            layer->qindex_table += 16;
      }
      layer->sample = layer->sample_gr;
      for (; layer->sb < frame->sblimit; layer->sb++) {
         if (*(layer->alloc)) {
            /* get sample values sample[0], sample[32], sample[64],
                 sample[51*32], sample[51*32+32], sample[51*32+64] */
            layer->qindex = layer->qindex_table[*(layer->alloc)];
            if (layer->qindex < 3) {
               /* read the grouped sample code */
               layer->qbits = qbits_table[layer->qindex];
               if (layer->qbits > 8) {
                  layer->qbits -= 8;
                  ALIGN_DATA_BYTE(SAMPLE_5, bit_shift_inc);
READ_SAMPLE_5:
                  GET_LSB;
                  *(layer->sample) = (AV1)DATA_BYTE << layer->qbits;
                  bit_shift_inc    = 8;
               }
               ALIGN_DATA_BYTE(SAMPLE_6, bit_shift_inc);
READ_SAMPLE_6:
               GET_LSB;
               *(layer->sample) |= (AV1)DATA_BYTE >> (8 - layer->qbits);
               bit_shift_inc     = layer->qbits;
#ifdef _MP2_TRACE_
               printf("     value: %d", *(layer->sample));
#endif
               *(layer->sample+((15+36)*32)) = *(layer->sample);
               requant_group(layer->sample, 
                             qsteps_table[layer->qindex],
                             *(layer->scalefactor));
               requant_group(layer->sample+((15+36)*32), 
                             qsteps_table[layer->qindex],
                             *(layer->scalefactor+3));
            }
            else {
               /* read the 3 separate samples */
#ifdef _MP2_TRACE_
               printf("    values:");
#endif
               for (layer->s = 0; layer->s < 3; layer->s++) {
                  layer->qbits = layer->qindex;
                  if (layer->qbits > 8) {
                     layer->qbits -= 8;
                     ALIGN_DATA_BYTE(SAMPLE_7, bit_shift_inc);
READ_SAMPLE_7:
                     GET_LSB;
                     *(layer->sample) = (AV1)DATA_BYTE << layer->qbits;
                     bit_shift_inc    = 8;
                  }
                  ALIGN_DATA_BYTE(SAMPLE_8, bit_shift_inc);
READ_SAMPLE_8:
                  GET_LSB;
                  *(layer->sample) |= (AV1)DATA_BYTE >> (8 - layer->qbits);
#ifdef _MP2_TRACE_
                  printf(" %d", *(layer->sample));
#endif
                  *(layer->sample+((15+36)*32)) = *(layer->sample);
                  layer->sample    += 32;    /* move to next sample */
                  bit_shift_inc     = layer->qbits;
               }
               layer->sample -= 3*32;        /* move back to first sample */
               requant_samples(layer->sample, 
                               layer->qindex, 
                               *(layer->scalefactor));
               requant_samples(layer->sample+((15+36)*32), 
                               layer->qindex, 
                               *(layer->scalefactor+3));
            }
#ifdef _MP2_TRACE_
            printf("\n  sample[0][%2hu][%2hu]=%d, %d, %d",
                   (u16)(layer->gr*3), (u16)layer->sb,
                   *(layer->sample), 
                   *(layer->sample+32), 
                   *(layer->sample+64));
            printf("\n  sample[1][%2hu][%2hu]=%d, %d, %d\n",
                   (u16)(layer->gr*3), (u16)layer->sb,
                   *(layer->sample+((15+36)*32)), 
                   *(layer->sample+((15+36)*32)+32), 
                   *(layer->sample+((15+36)*32)+64));
#endif
         }
         layer->alloc       += 2;   /* move to next sub-band */
         layer->scalefactor += 6;   /* move to next sub-band */
         layer->sample++;           /* move to next sub-band */
         if (layer->sb == *(layer->qindex_table))
            layer->qindex_table += 16;
      }
   }

   /***** CREATE AUDIO SAMPLES *****/
   if (!layer->audio_buf)
      goto LAYER2_read_DONE;
   audio_buf     = layer->audio_buf;
   layer->sample = layer->sample_array0;
   for (layer->ch = 0; layer->ch < frame->num_ch; layer->ch++) {
      for (layer->s = 0; layer->s < 36; layer->s++) {
         dct32(layer->sample);
         synth_filter(layer->sample, audio);
#ifndef _MP2_TRACE_
         for (bit_shift_inc = 0; bit_shift_inc < 32; bit_shift_inc++) {
            *audio_buf = audio[bit_shift_inc];
            audio_buf += frame->num_ch;
         }
#else
         printf("SYNTH (ch=%hu, i=%hu):\n  DCT32:\n",
                (u16)layer->ch, (u16)layer->s);
         for (bit_shift_inc = 0; bit_shift_inc < 32; bit_shift_inc += 4) {
            printf("  0x%08X  0x%08X  0x%08X  0x%08X\n",
                   layer->sample[bit_shift_inc],
                   layer->sample[bit_shift_inc+1],
                   layer->sample[bit_shift_inc+2],
                   layer->sample[bit_shift_inc+3]);
         }
         printf("  AUDIO:\n");
         for (bit_shift_inc = 0; bit_shift_inc < 32; bit_shift_inc += 4) {
            printf("  %6hd  %6hd  %6hd  %6hd\n",
                   audio[bit_shift_inc],
                   audio[bit_shift_inc+1],
                   audio[bit_shift_inc+2],
                   audio[bit_shift_inc+3]);
         }
#endif
         layer->sample += 32; /* move to the next group of samples */
      }
      layer->sample += 15*32; /* move to the channel */
      audio_buf     -= 36*32 * (u32)frame->num_ch;
      audio_buf++;
   }
   DCT32_DONE;
#ifndef _MP2_TRACE_
# ifndef _SPEED_TEST_
   if (frame->pes_pts) {
      CLK_VAL cur_pts;
   
      cur_pts = MPEG2_prog_clk(frame->offset) +
                (CLK_VAL)((MPEG2_SYS_CLK_FREQ*SOUND_samples_left())/frame->sample_rate);
      *(frame->a_sync) = cur_pts - frame->pes_pts;
      frame->pes_pts   = 0;
   }
# endif
   SOUND_play();
#endif

   /* store the samples needed for the next frame */
   memcpy(layer->prev_sample_array0,
          layer->sample_array0 + ((36-15)*32), 
          (sizeof(AV1)*15*32));
   memcpy(layer->prev_sample_array1,
          layer->sample_array1 + ((36-15)*32), 
          (sizeof(AV1)*15*32));

LAYER2_read_DONE:
   frame->state = LAYER2_read_END_STATE;
#ifdef _MP2_TRACE_
   printf("END DECODE LAYER 2 (FRAME %d)\n", frame_cnt);
   frame_cnt++;
#endif

#undef ALLOC_1
#undef ALLOC_2
#undef SCFSI
#undef SCALEFACTOR_1
#undef SCALEFACTOR_2
#undef SCALEFACTOR_3
#undef SAMPLE_1
#undef SAMPLE_2
#undef SAMPLE_3
#undef SAMPLE_4
#undef SAMPLE_5
#undef SAMPLE_6
#undef SAMPLE_7
#undef SAMPLE_8
}


static void requant_samples(AV1 *sample, u8 qbits, u8 scalefactor)
{
   AV2 add2, sample2;
   s32 add1, sf, i;

   qbits--;   
   add1   = 1 - (1 << qbits);
   sf     = sf_table1[qbits - 1][scalefactor % 3];
   qbits += scalefactor / 3;
   add2   = (AV2)1 << (qbits - 1);
   
   for (i = 0; i < 3; i++) {
      sample2   = (AV2)(*sample + add1);
      sample2  *= (AV2)sf;
      sample2  += add2;
      sample2 >>= qbits;
      *sample   = (AV1)sample2;
      sample   += 32;
   }
}

static void requant_group(AV1 *sample, u8 qsteps, u8 scalefactor)
{
   s32 i, add, code, sample1,
       sub = (s32)qsteps >> 1,
       sf  = sf_table2[qsteps >> 2][scalefactor % 3];

   scalefactor /= 3;
   add          = (scalefactor) ? (1 << (scalefactor - 1)) : 0;
   code         = (s32)*sample;
   for (i = 0; i < 3; i++) {
      sample1  = code % (s32)qsteps;
      code    /= (s32)qsteps;
      sample1 -= sub;
      sample1 *= sf;
      sample1 += add;
      *sample  = (AV1)(sample1 >> scalefactor);
      sample  += 32;
   }
}

#undef DATA_BYTE
#undef BIT_SHIFT_BYTE
#undef DATA_PTR
#undef END_DATA_PTR
#undef PARSER_STATE
#undef STOP_PARSING
#undef SF1_SCALE
#undef SF1_CONST
#undef SF1_ROW
#undef SF2_ROW

